/*
	File:			makelinks.c

	Function:		Creates soft links from a 'make' directory to the 
					original source files.

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1996-1999, Andrew Willmott
*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

typedef int bool;
#define true 1
#define false 0
#define MAX_PATHS 512

typedef struct
{
	int		numPaths;
	char	*path[MAX_PATHS];
} PathsArray;

void PrintUsage(char *programName)
{
	fprintf(stderr,
		"Usage: %s [-clean][-dep NAME][-paths] path1 path2 ... -files file1 file2 ...\n",
			programName);
	fprintf(stderr,
		"	 -clean removes the old symbolic links: no paths need be\n"
		"	 specified with this option.\n");
	fprintf(stderr,
		"	 -dep generates a list of NAME: file dependencies suitable for \n"
		"    a makefile.\n");
	fprintf(stderr,
	"Purpose: Searches for the specified files in the specified paths,\n"
	"starting from 'path1'. When it finds a file, it makes a symlink to \n"
	"it in the current directory.\n");
}

void CleanOldLink(char *filename)
{
	struct stat		buffer;

/* If the file exists in the current directory, and is a link, delete it. */

	if (lstat(filename, &buffer) == 0 && (buffer.st_mode & S_IFMT) == S_IFLNK)
		if (unlink(filename) != 0)
			perror("unlink");
}

void MakeFileLink(char *filename, PathsArray *paths, 
				  char *fDependName, bool fGetPaths)
{
	char			tempPath[1024];	
	struct stat		buffer;
	int				i;
		
	/* Search through the paths in order... */

	for (i = 0; i < paths->numPaths; i++)
	{
		strcpy(tempPath, paths->path[i]);
		strcat(tempPath, "/");
		strcat(tempPath, filename);
		
		if (stat(tempPath, &buffer) == 0)
		{

			if (fDependName)
			{
				/* print dependency */
				printf("%s: %s\n", fDependName, tempPath);
			}
			/* Create link to file if it's not in the current directory */
			else if (strcmp(paths->path[i], ".") != 0)	
				if (symlink(tempPath, filename) != 0 && errno != EEXIST)
					perror("symlink");
			return;
		}
	}
	
	if (!fDependName)
		fprintf(stderr, "Can't find file: %s\n", filename);
	return;
}


int main(int argc, char **argv)
{
	PathsArray	paths;
	int			currentArg = 1;
	bool		fGetPaths = true;
	bool		fClean = false;
	char 		*fDependName = 0;

	if (argc < 2)
	{
		fprintf(stderr, "Insufficient arguments\n");
		PrintUsage(argv[0]);
		exit(1);
	}
	
	paths.numPaths = 0;
		
	while (fGetPaths && currentArg < argc)	
	{
		if (*argv[currentArg] == '-')
		{
			if (strcmp(argv[currentArg], "-paths") == 0)
				; /* ignore it */
			else if (strcmp(argv[currentArg], "-files") == 0)
				fGetPaths = false;
			else if (strcmp(argv[currentArg], "-clean") == 0)
				fClean = true;
			else if (strcmp(argv[currentArg], "-dep") == 0)
			{
				currentArg++;
				fDependName = argv[currentArg];
				
				paths.path[paths.numPaths] = "/usr/lib";
				paths.numPaths++;
				paths.path[paths.numPaths] = "/usr/lib32";
				paths.numPaths++;
			}
			else if (argv[currentArg][1] == 'L')
			{
				paths.path[paths.numPaths] = argv[currentArg] + 2;
				paths.numPaths++;
			}	
			else if (argv[currentArg][1] == 'I')
			{
				paths.path[paths.numPaths] = argv[currentArg] + 2;
				paths.numPaths++;
			}	
			else if (argv[currentArg][1] == 'l')
			{
				char libfile[1024];

				sprintf(libfile, "lib%s.a", argv[currentArg] + 2);
				MakeFileLink(libfile, &paths, fDependName, fGetPaths);
				sprintf(libfile, "lib%s.so", argv[currentArg] + 2);
				MakeFileLink(libfile, &paths, fDependName, fGetPaths);
			}	
			else
			{
				fprintf(stderr, "Unrecognized flag: %s\n", argv[currentArg]);
				PrintUsage(argv[0]);
				exit(1);
			}
		}
		else
		{
			if (paths.numPaths >= MAX_PATHS)
				fprintf(stderr, 
						"*** ERROR: Can't handle more than %d paths: "
						"ignoring %s\n", MAX_PATHS, argv[currentArg]);
			else
			{
				paths.path[paths.numPaths] = argv[currentArg];
				paths.numPaths++;
			}			
		}
		
		currentArg++;
	}
	
	while (currentArg < argc)
	{
		if (*argv[currentArg] == '-')
		{
			fprintf(stderr, "Unexpected flag: %s\n", argv[currentArg]);
			PrintUsage(argv[0]);
			exit(1);
		}
				
		if (!fDependName) 
			CleanOldLink(argv[currentArg]);
		if (!fClean)
			MakeFileLink(argv[currentArg], &paths, fDependName, fGetPaths);
					
		currentArg++;
	}
	
	return(0);
}
