/*
	File:		Expr.cc
 
	Function:	
 
	Authors:	Paul Heckbert, Andrew Willmott
 
	Notes:		
*/

#include "cl/Expr.h"

/*
 * expr_eval: expression evaluator - converts ascii string to floating point
 * Works by top-down predictive parsing.
 * Most of the routines gobble characters and advance global string pointer s.
 * Sets global expr_err if an error occurs.
 *
 * supports: parentheses, % for mod, ^ for pow, elementary functions,
 * constants pi and e, variable base constants
 *
 * Paul Heckbert	18 April 1988
 * converted to ANSI C, ajw 1999
 */


#include <ctype.h>
#include <math.h>
#include <limits.h>

inline Double DegsToRads(Double degrees)	///< Degrees to radians conversion...
{ return(degrees * (M_PI / 180.0)); }
inline Double RadsToDegs(Double rads)		///< Radians to degrees conversion...
{ return(rads * (180.0 / M_PI)); }

static Double expr();
static Double term();
static Double factor();
static Double signednumber();
static Double number();
static Double paren();
static Double posconst();
static Double expt(Int a, Int n);
static void paren2(Double *x, Double *y);
static Int eq(Int n, const Char *a, const Char *b);
static void error(const Char *s, Int len, const Char *err);
static void prints(Int n, const Char *s);
static Int digit(Int c);

static const Char *s0, *s;
Int expr_error;

inline void space()
{ for (; isspace(*s); s++); };

#ifdef MAIN
main(Int ac, Char **av)
{
    Double x;

    if (ac != 2) exit(1);
    x = expr_eval(av[1]);
    printf("%g\n", x);
}
#endif

Int expr_eval_int(StrConst str)
{
    Double x;

    x = expr_eval(str);
    /* do unsigned Double to signed Int conversion: */

	if (x > INT_MAX)
		return(Int(x + 2.0 * INT_MIN));
	else
		return(Int(x));
}

Long expr_eval_long(StrConst str)
{
    Double x;

    x = expr_eval(str);
    /* do unsigned Double to signed Long conversion: */
	if (x > LONG_MAX)
		return(Long(x + 2.0 * LONG_MIN));
	else
		return(Long(x));
}

Double expr_eval(StrConst str)
{
    Double x;

    s0 = s = str;
    expr_error = EXPR_GOOD;
    x = expr();
    if (*s)
    {
        error(s, 1, "garbage in expression");
        expr_error = s == s0 ? EXPR_BAD : EXPR_SOSO;
    }
    return x;
}

static Double expr()
{
    Double x;

    for (x = term(); ; )
    {
        space();
        switch (*s)
        {
        case '+':
            s++;
            x += term(); break;
        case '-': s++; x -= term(); break;
        default: return x;
        }
    }
}

static Double term()
{
    Double x, y;

    for (x = factor(); ; )
    {
        space();
        switch (*s)
        {
        case '*':
            s++;
            x *= factor(); break;
        case '/': s++; x /= factor(); break;
        case '%': s++; y = factor(); x = x - floor(x / y) * y; break;
        default: return x;
        }
    }
}

static Double factor()
{
    Double x;

    for (x = signednumber(); ; )
    {
        space();
        switch (*s)
        {
        case '^':
            s++;
            return pow(x, factor());  	/* right-associative */
        default: return x;
        }
    }
}

static Double signednumber()
{
    space();
    switch (*s)
    {
    case '-':
        s++;
        return - signednumber();
    case '+': s++; return signednumber();
    default: return number();
    }
}

static Double number()
{
    const Char *func;
    Int n;
    Double x, y;

    space();
    if (isdigit(*s) || *s == '.') return posconst();
    if (*s == '(') return paren();

    if (isalpha(*s))
    {
        func = s;
        for (s++; isalpha(*s) || isdigit(*s); s++);
        n = s - func;  	/* length of funcname */

        if (eq(n, func, "pi"))	return M_PI;
        if (eq(n, func, "e"))	return exp(1.);

        if (eq(n, func, "sqrt"))	return sqrt(paren());
        if (eq(n, func, "exp"))	return exp(paren());
        if (eq(n, func, "log"))	return log(paren());
        if (eq(n, func, "pow"))
        {
            paren2(&x, &y);
            return pow(x, y);
        }

        if (eq(n, func, "sin"))	return sin(paren());
        if (eq(n, func, "cos"))	return cos(paren());
        if (eq(n, func, "tan"))	return tan(paren());
        if (eq(n, func, "asin"))	return asin(paren());
        if (eq(n, func, "acos"))	return acos(paren());
        if (eq(n, func, "atan"))	return atan(paren());
        if (eq(n, func, "atan2"))
        {
            paren2(&x, &y);
            return atan2(x, y);
        }

        if (eq(n, func, "sind"))	return sin(DegsToRads(paren()));
        if (eq(n, func, "cosd"))	return cos(DegsToRads(paren()));
        if (eq(n, func, "tand"))	return tan(DegsToRads(paren()));
        if (eq(n, func, "dasin"))	return RadsToDegs(asin(paren()));
        if (eq(n, func, "dacos"))	return RadsToDegs(acos(paren()));
        if (eq(n, func, "datan"))	return RadsToDegs(atan(paren()));
        if (eq(n, func, "datan2"))
        {
            paren2(&x, &y);
            return RadsToDegs(atan2(x, y));
        }

        if (eq(n, func, "floor"))	return floor(paren());
        if (eq(n, func, "ceil"))	return ceil(paren());

        error(func, n, "bad numerical expression");
        return 0.;
    }

    error(s, 1, "syntax error");
    return 0.;
}

/* paren: '(' expr ')' */

static Double paren()
{
    Double x;

    space();
    if (*s != '(') error(s, 1, "expected '('");
    s++;
    x = expr();
    space();
    if (*s != ')') error(s, 1, "expected ')'");
    s++;
    return x;
}

/* paren2: '(' expr ',' expr ')' */

static void paren2(Double *x, Double *y)
{
    space();
    if (*s != '(') error(s, 1, "expected '('");
    s++;
    *x = expr();
    space();
    if (*s != ',') error(s, 1, "expected ','");
    s++;
    *y = expr();
    space();
    if (*s != ')') error(s, 1, "expected ')'");
    s++;
}

/*
 * posconst: given a string beginning at s, return floating point value.
 * like atof but it uses and modifies the global ptr s
 */

static Double posconst()
{
    Int base, exp, pos, d;
    Double x, y;

    space();
    if (*s == '0')
    {		/* change base: 10 = 012 = 0xa = 0b2:1010 */
        s++;
        switch (*s)
        {
        case 'b':
            s++;
            for (base = 0; isdigit(*s); s++)
                base = base * 10 + *s - '0';  	/* base is in base 10! */
            if (*s != ':') error(s, 1, "expecting ':'");
            s++;
            break;
        case 'x':
            s++;
            base = 16;
            break;
        case 't':
            s++;
            base = 10;
            break;
        case '.':
            base = 10;
            break;  		/* a float, e.g.: 0.123 */
        default:
            base = 8;
            break;
        }
    }
    else base = 10;

    x = 0.;
    for (; d = digit(*s), d >= 0 && d < base; s++)
        x = x * base + d;
    if (*s == '.')
    {
        s++;
        for (y = 1.; d = digit(*s), d >= 0 && d < base; s++)
        {
            x = x * base + d;  		/* fraction is in variable base */
            y *= base;
        }
        x /= y;
    }
    if (*s == 'e' || *s == 'E')
    {
        s++;
        if (*s == '-')
        {
            s++;
            pos = 0;
        }
        else if (*s == '+')
        {
            s++;
            pos = 1;
        }
        else pos = 1;
        for (exp = 0; isdigit(*s); s++)
            exp = exp * 10 + *s - '0';  	/* exponent is in base 10 */
        y = expt(base, exp);
        if (pos) x *= y;
        else x /= y;
    }
    return x;
}

static Int digit(Int c)
{
    return isdigit(c) ? c - '0' :
           c >= 'a' && c <= 'z' ? c - 'a' + 10 : c >= 'A' && c <= 'Z' ? c - 'A' + 10 : -1;
}

/* expt: a^n for n>=0 */

static Double expt(Int a, Int n)
{
    Double t, x;

    if (n < 0)
    {
        fprintf(stderr, "expt: can't do negative exponents\n");
        return 1.;
    }
    if (n == 0) return 1.;
    for (t = a, x = 1.; n > 0; n >>= 1)
    {
        if (n&1) x *= t;
        t *= t;
    }
    return x;
}

/* eq: test equality of string a, of length n, with null-terminated string b */

static Int eq(Int n, const Char *a, const Char *b)
{
    Char c, *nca = (Char *) a;
    Int ret;

    c = a[n];
    nca[n] = 0;
    ret = (strcmp(a, b) == 0);
    nca[n] = c;
    return ret;
}

static void error(const Char *s, Int len, const Char *err)
{
//    if (*s == 0) s[len] = 0;  	/* just in case */
    printf("expr: %s: ", err);
    prints(s - s0, s0);
    printf("[");
    prints(len, s);
    printf("]");
    prints(s + strlen(s) - s0 - len, s + len);
    printf("\n");
    if (expr_error != EXPR_BAD)
        expr_error = s == s0 ? EXPR_BAD : EXPR_SOSO;
}

/* prints: print string s of length n */

static void prints(Int n, const Char *s)
{
    Char c, *ncs = (Char *) s;

    c = s[n];
    ncs[n] = 0;
    printf("%s", s);
    ncs[n] = c;
}
