/*
	File:		EPSPlot.cc

	Function:	Provides 2D plotter which outputs EPS files.

	Author:		Andrew Willmott

	Notes:		Some of this is originally from Paul Heckbert's libplo.a
*/

#include "gcl/EPSPlot.h"
#include "gcl/Geometry.h"
#include <string.h>

// --- eps header -------------------------------------------------------------

static Char header[] = 
"%%!PS-Adobe-3.0 EPSF-3.0\n"
"%%%%BoundingBox: %d %d %d %d\n"
"%%%%Title: %s\n"
"%%%%Creator: PSPlot class\n";

// --- little ps library  -----------------------------------------------------

static Char definitions[] = "\
/m {moveto} def\n\
/d {lineto} def\n\
/t {moveto show} def\n\
/p {stroke} def\n\
/dot {newpath 0 360 arc closepath fill} def\n\
/cir {newpath 0 360 arc closepath stroke} def\n\
\n\n";


// --- EPSPlot class ----------------------------------------------------------


EPSPlot::EPSPlot() : 
	file(0),
	drawn(false),
	font("Helvetica"),
	width(1),
	scale(144),		// default 2" x 2"
	level2(false)
{
}

Void EPSPlot::Open(const Char *filename)
{
    if (strcmp(filename, ".") == 0)
    	file = stdout;
    else if ((file = fopen(filename, "w")) == NULL)
    {
		fprintf(stderr, "(EPSPlot::Open) Can't write %s\n", filename);
		return;
    }
    
    fprintf(file, header, 0, 0, (int) scale, (int) scale, "plot");
    fputs(definitions, file);
	
	if (level2)
	{
		// draw border
		fprintf(file, "0.1 setlinewidth\n");
		fprintf(file, "%g %g %g %g rectstroke\n", 0.0, 0.0, scale, scale);
	
		// make sure everything outside viewport is clipped.
		fprintf(file, "%g %g %g %g rectclip\n", 0.0, 0.0, scale, scale);
	}
	    
	// scale down so coord. sys. is [0,1].
	
	fprintf(file, "%g %g scale\n", scale, scale);
}

Void EPSPlot::Close()
{
    if (drawn)
    	Stroke();
    fprintf(file, "showpage\n");
    fclose(file);
}

// --- Drawing commands -------------------------------------------------------

Void EPSPlot::MoveTo(GCLReal x, GCLReal y)
{
    if (drawn)
    	Stroke();
    fprintf(file, "%.4g %.4g m\n", x, y);
}

Void EPSPlot::LineTo(GCLReal x, GCLReal y)
{
    drawn = 1;
    fprintf(file, "%.4g %.4g d\n", x, y);
}

Void EPSPlot::Stroke()
{
    fprintf(file, "p\n");
    drawn = 0;
}

Void EPSPlot::SetWidth(GCLReal width)
{
	if (drawn)
		Stroke();
	fprintf(file, "%g setlinewidth\n", width);
}

Void EPSPlot::LineStyle(Int style)
{
    if (drawn)
    	Stroke();
    switch (style)
    {
	case lineDashed:
		fprintf(file, "[.06 .04] 0 setdash\n");
		break;
	case lineDotted:
		fprintf(file, "[.015 .03] 0 setdash\n");
		break;
	default:
	    fprintf(stderr, "(EPSPlot::LineStyle) don't know style %d\n", style);
	    /* fall through */
	case lineSolid:
		fprintf(file, "[] 0 setdash\n");
		break;
    }
}

Void EPSPlot::SetColour(const Colour &clr)
{
	if (drawn)
		Stroke();

	if (level2)
	{
		fprintf(file, "%g %g %g setrgbcolor\n", 
			Double(clr[0]), Double(clr[1]), Double(clr[2]));
	}
	else
	{
		Double level = dot(cRGBToLum, clr);
		fprintf(file, "%g setgray\n", level);
	}
}

Void EPSPlot::Dot(GCLReal x, GCLReal y, GCLReal r)
{
    if (drawn)
    	Stroke();
    fprintf(file, "%.4g %.4g %.4g dot\n", x, y, r);
}

Void EPSPlot::Circle(GCLReal x, GCLReal y, GCLReal r)
{
    if (drawn)
    	Stroke();
    fprintf(file, "%.4g %.4g %.4g cir\n", x, y, r);
}

Void EPSPlot::DrawText(GCLReal x, GCLReal y, Char *string)
{
    fprintf(file, "(%s) %.4g %.4g t\n", string, x, y);
}

Void EPSPlot::SetOutputSize(GCLReal size)
{
	scale = size;
}

Void EPSPlot::SetPort(GCLReal left, GCLReal bottom, GCLReal right, GCLReal top,
					  Bool centred)
{
	Scale(1.0 / (right - left), 1.0 / (top - bottom));
	if (centred)
		Translate((right - left) / 2.0, (top - bottom) / 2.0);
	else
		Translate(-left, -bottom);
}

Void EPSPlot::Push()
{
    if (drawn)
    	Stroke();
    fprintf(file, "gsave\n");
}

Void EPSPlot::Pop()
{
    if (drawn)
    	Stroke();
    fprintf(file, "grestore\n");
}

Void EPSPlot::Translate(GCLReal tx, GCLReal ty)
{
    fprintf(file, "%.4g %.4g translate\n", tx, ty);
}

Void EPSPlot::Scale(GCLReal sx, GCLReal sy)
{
    fprintf(file, "%.4g %.4g scale\n", sx, sy);
}

Void EPSPlot::Rotate(GCLReal angle)
{
    fprintf(file, "%.4g rotate\n", RadsToDegs(angle));
}

Void EPSPlot::Comment(StrConst string)
{
    fprintf(file, "%% %s\n", string);
}


