/*
	File:			EPSRenderer.cc

	Function:		Implements EPSRenderer.h

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

*/

#include <iostream.h>
#include "gcl/EPSRenderer.h"

EPSRenderer::EPSRenderer() : 
	Renderer(),
	matrixStack(1, 32),
	start(true),
	style(renNone),
	itsFilename("out.eps")
{
	matrixStack.Top() = vl_I;
	itsPlot.SetOutputSize(192);	
	// 192 points: this gets us same size as 400x400 picture at 150dpi in frame
}

Void EPSRenderer::ViewPoint(const HPoint &hp, Bool draw)
{
	Point p = proj(hp);

	if (draw)
		itsPlot.LineTo(p[0], p[1]);
	else
		itsPlot.MoveTo(p[0], p[1]);
}

Void EPSRenderer::Attach(StrConst epsFilename)
{
	itsFilename = epsFilename;
}

Void EPSRenderer::MakeCurrent()
{
}

Void EPSRenderer::Show()
{
	itsPlot.Close();
}


// --- Renderer Drawing Operators ---------------------------------------------


Renderer &EPSRenderer::SetColour(const Colour &c)
{
	itsPlot.SetColour(c);
	return(SELF);
}

Renderer &EPSRenderer::SetPoint(const Point &p)
{
	HPoint	hp(p, 1.0);

	hp = xform(matrixStack.Top(), hp);

	ClipPoint(hp, !start);

	if (start)
	{
		firstPoint = p;
		start = false;
	}
		
	return(SELF);
}

Renderer &EPSRenderer::SetCoord(const Coord &c)
{
	return(SetPoint(Vector(c, 0.0)));
}

Renderer &EPSRenderer::SetNormal(const Vector &n)
{
	return(SELF);
}

Renderer &EPSRenderer::SetTransform(const Transform &t)
{
	matrixStack.Top() *= t;	
	return(SELF);
}

Renderer &EPSRenderer::SetCamera(const Camera &c)
{
	matrixStack.Clear();
	matrixStack.Push(c.ProjMatrix() * c.ModelMatrix());
	return(SELF);
}

Renderer &EPSRenderer::Pop()
{
	matrixStack.Pop();
	return(SELF);
}

Renderer &EPSRenderer::Push()
{
	matrixStack.Push(matrixStack.Top());	
	return(SELF);
}

Renderer &EPSRenderer::Clear()
{
	itsPlot.Open(itsFilename);	
	itsPlot.SetPort(-1, -1, 1, 1, true);
	itsPlot.SetWidth(0.0005);

	return(SELF);
}

Renderer &EPSRenderer::Begin(RenderStyle newStyle)
{
	Assert(style == renNone, "Unmatched EPSRenderer::Begin().");
	style = newStyle;
	start = true;
	return(SELF);
}

Renderer &EPSRenderer::End()
{
	Assert(style != renNone, "Unmatched EPSRenderer::End().");

	if (style == renPoly || style == renLineLoop)
		// close loop
		SetPoint(firstPoint);

	style = renNone;
	return(SELF);
}

Void EPSRenderer::Print(ostream &s)
{
#ifdef BROKEN 
	s << "EPS Renderer\n  style = " << (int) style
	  << " matrix = " << matrixStack.Top()
	  << " start = " << start << endl;	
#endif
}

