/*
	File:		GCLApp.cc

	Function:	

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/GCLApp.h"
#include "gcl/Avars.h"
#include "gcl/Readers.h"
#include <stdlib.h>

#define DBG_COUT if (0) cerr


// --- GCLApp --------------------------------------------------------------


/** \class GCLApp

    Provides useful routines for any GCL application that must deal with
    a scene and avars.

 */

Char			**GCLApp::files;
Int				GCLApp::numFiles;
StrConstArray	GCLApp::avarNames;
ScalarList		GCLApp::avarVals;

GCLApp::GCLApp() : 
#ifndef GCL_NO_GL
    itsScenePane(0),
#endif
    itsCamera(0),
    frame(1),
    time(0.0),
    fps(24.0),
    anim(0),
    avarList(0)
{
}

Void GCLApp::ReadScene(FileName &sceneFile)
{
	scene = SceneReader::Load(sceneFile);
	if (!scene)
		return;

	if (zUp)
		scene->Apply(Rotation(vl_x, vl_halfPi));
	else if (zDown)
		scene->Apply(Rotation(vl_x, -vl_halfPi));
	
	if (!noScale)
		scene->Normalise();
}

Void GCLApp::GetFileArgs(Int argc, Char *argv[])
{
	// Save away a pointer to the list of file names
	// I've checked the source of libarg.c -- it
	// doesn't free anything, so this is safe.

	files = argv;
	numFiles = argc;
}

Void GCLApp::GetAvarArgs(Int argc, Char *argv[])
{
	Int		i;
	String	suffix;

	for (i = 0; i < argc; )
	{
		StrConst avarName = argv[i++];
		
		DBG_COUT << "looking at " << avarName << " suff = " << avarName.Suffix(2) << endl;

		suffix = avarName.Suffix(2);
		if (suffix == "_c" || suffix == "_v" || suffix == "_q")
		{
			String stub, newAvar;

			stub = avarName.Prefix(-2);

			DBG_COUT << "got vector type " << stub << endl;

			newAvar.Printf("%s_x", stub.CString());
			avarNames.Append(newAvar.Detach());
			avarVals.Append(atof(argv[i++]));
			newAvar.Printf("%s_y", stub.CString());
			avarNames.Append(newAvar.Detach());
			avarVals.Append(atof(argv[i++]));
			if (suffix[1] == 'v' || suffix[1] == 'q')
			{
				newAvar.Printf("%s_z", stub.CString());
				avarNames.Append(newAvar.Detach());
				avarVals.Append(atof(argv[i++]));
			}
			if (suffix[1] == 'q')
			{
				newAvar.Printf("%s_w", stub.CString());
				avarNames.Append(newAvar.Detach());
				avarVals.Append(atof(argv[i++]));
			}
		}
		else
		{
			avarNames.Append(avarName);
			avarVals.Append(atof(argv[i++]));
		}		
	}

	if (i != argc)
	{
		cerr << "Error: wrong number of avar arguments." << endl;
		exit(1);
	}
}

Void GCLApp::SetAvars(scScenePtr scene)
{
	Int				i;
	CreateAvarList	theAvarList;

	frame = -1;
		
	scene->ApplyAction(theAvarList);

#ifdef DEBUG
	cout << "set avars:" << endl;
	for (i = 0; i < avarNames.NumItems(); i++)
		cout << "avar " << avarNames[i] << " = " << avarVals[i] << endl;
	cout << "scene avars:" << endl;
	for (i = 0; i < theAvarList.avarList->NumItems(); i++)
		cout << theAvarList.avarList->Item(i).name << " = " 
			<< theAvarList.avarList->Item(i).value << endl;
#endif

	scene->Set(theAvarList.avarList);
	
	for (i = 0; i < avarNames.NumItems(); i++)
	{
		StrConst avarName = avarNames[i];

		if (avarName == "frame")
			frame = (Int) avarVals[i];
		else if (avarName == "time")
			time = avarVals[i];
		else if (itsCamera->CheckAvar(avarName, avarVals[i]))
			;
		else if (!theAvarList.avarList->SetAvar(avarNames[i], avarVals[i]))
			cerr << "No such avar in scene: " << avarNames[i] 
				 << " (ignoring) " << endl;
	}
	
	itsCamera->SetupFromParams();
}

Void GCLApp::SetAvarsFromAnim(scScenePtr scene)
{
	if (!avarList)
	{
		avarList = (scAvarList*) scene->FindFirst(aAvarList);

		if (!avarList)
		{
			CreateAvarList	theAvarList;

			cout << "creating avar list" << endl;
			scene->ApplyAction(theAvarList);
			scene->Set(theAvarList.avarList);
			avarList = theAvarList.avarList;
		}
	}
	
	time = (frame - 1.0) / fps;
	if (anim)
		anim->SetAvarList(avarList, itsCamera, time);
}

ArgForm *GCLApp::GetOptionList()
{
#ifdef UNFINISHED
	return(arg_to_form(0,
		"-bgr %F", &grey,
			"Set background grey level",
		"-bgc %F %F %F", clr, clr + 1, clr + 2,	
			"Set colour background",
		"-defClr %F %F %F", defClr, defClr + 1, defClr + 2,	
			"Set default model colour",
		"-noScale", ARG_FLAG(&noScale),
			"Don't rescale scene.",
//		"-wire", ARG_FLAG(&wireFrame),
//			"Draw wire-frame.",
		"-zup", ARG_FLAG(&zUp),
			"Z axis is up (default is Y)",
		"-zdown", ARG_FLAG(&zDown),
			"Z axis is down",
//		"-noLight", ARG_FLAG(&noLight),
//			"no lighting",
//		"-showBF", ARG_FLAG(&showBF),
//			"show back faces of polygons",

//		"-anim %S",		&animFile,					"specify animation file",
		"-fps %F",		&fps,						"set frames per sec [default 24]",
//		"-start %d", 	&startFrame,				"start frame [1]",
//		"-end %d", 		&endFrame,					"end frame [24]",
//		"-step %d", 	&frameStep,					"frame step [1]",

		"-set", ARG_SUBR(GetAvarArgs),
			"Set avar, e.g. -set light 0.5 height 0.2",

		// Output options		
//		"-size %d", &size,
//			"Set output size",
		"-xysize %d %d", &xsize, &ysize,
			"Set output size",
		0)
	);
#else
	return(0);
#endif
}
