/*
	File:		SceneObjects.cc

	Function:	See header file

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/SceneObjects.h"
#include "gcl/VecUtil.h"
#include "gcl/GCLConfig.h"

// --- Polygon methods --------------------------------------------------------
    
StrConst scPoly::Label() const
{
	return("mesh");
}

IndexList *gIdentityList = 0;


/*
	Current defaults:

	point indexes = [0, 1, 2, 3...]
	normal, texture indexes: point indexes
	colour indexes: normal indexes
*/

Void scPoly::Draw(Renderer &r, SLContext *context)
// draw one or more polygons
{
	Int 		i, j;
	Int			f, faces, faceS, faceE;
	Vector		normal;
	Colour		polyClr;
	PointList	*points;
	NormalList	*normals;
	ColourList	*colours;
	CoordList	*coords;
	Texture		*texture;
	IndexList	*indexes;
	IndexList	*clrIndexes;
	IndexList	*normIndexes;
	IndexList	*coordIndexes;
	IndexList	*faceIndexes;
	scMeshType	*meshType;
	RenderStyle	meshStyle;
	
	points = SC_GET(Points);
	colours = SC_GET(Colours);
	normals = SC_GET(Normals);
	texture = SC_GET(Texture);

	if (!gIdentityList)
	{
		gIdentityList = new IndexList(256);
		for (i = 0; i < gIdentityList->NumItems(); i++)
			gIdentityList->Item(i) = i;
	}

	indexes = SC_GET(PointIndexes);
	faceIndexes = SC_GET(FaceIndexes);
	if (meshType = SC_GET(MeshType))
		meshStyle = meshType->itsType;
	else
		meshStyle = sRenderStyle;

	if (!indexes)
	{
		indexes = gIdentityList;
		Assert(points->NumItems() <= 256, "polygon too large");
		// note: we are relying on Array not changing the current items
		// array when we resize down and then back up again. 
		// (This is currently safe: the items array only changes if the array
		// size is increased.)
		indexes->SetSize(points->NumItems());
	}

	normIndexes = SC_GET(NormalIndexes);
	if (!normIndexes)
		normIndexes = indexes;

	if (colours)
	{
		clrIndexes = SC_GET(ColourIndexes);
		if (!clrIndexes)
			clrIndexes = normIndexes;
	}
	else
	{
		if (SC_GET(Colour))
			polyClr = *(SC_GET(Colour));
		else if (SC_GET(Emittance))
			polyClr = *(SC_GET(Emittance));
		else
			polyClr[0] = -1.0;
	}
			
	if (texture)
	{
		// texture map
		coords = SC_GET(Coords);
		if (coords)
		{
			r.SetTexture(texture->image);
			coordIndexes = SC_GET(CoordIndexes);
			if (!coordIndexes)
				coordIndexes = gIdentityList;
		}
	}
	else 
		coords = 0;
		
	if (faceIndexes)
	{

		if (faceIndexes->NumItems() < 2)
		{
			// XXX empty face index list
			// should modify face begin & end routines to cull
			// such empty lists; requires object->Remove() method?
//			_Warning("Empty face index list");
			return;
		}
		
		faceS = faceIndexes->Item(0);
		faceE = faceIndexes->Item(1);
		faces = faceIndexes->NumItems() - 1;
		f = 2;
	}
	else
	{
		faceS = 0;
		faceE = indexes->NumItems();
		faces = 1;
	}
			
	while (true)
	{
		r.Begin(meshStyle);

		if (!normals)
		{
			CalcTriAreaNormal((*points)[(*indexes)[faceS]], 
				(*points)[(*indexes)[faceS + 1]], 
				(*points)[(*indexes)[faceS + 2]], 
				normal
			);
			r.N(normal);
		}
		
		if (!colours && polyClr[0] >= 0.0)
			r.SetColour(polyClr);
				
		for (i = faceS; i < faceE; i++)
		{
			if (colours)
			{
				j = (*clrIndexes)[i];
				r.SetColour((*colours)[j]);
			}
			if (coords)
			{
				j = (*coordIndexes)[i];
				r.SetTexCoord((*coords)[j]);
			}
			if (normals)
			{
				j = (*normIndexes)[i];
				r.SetNormal((*normals)[j]);
			}
				
			j = (*indexes)[i];
			r.SetPoint((*points)[j]);
		}
		r.End();

		if (faces <= 1) break;
		faces--;
		faceS = faceE;
		faceE = faceIndexes->Item(f++);
	}

	if (coords)
		r.SetTexture(0);
}

Void scPoly::DecimateSelf(Decimator &dec)
{
	Int			i, j, n, idx[32], changed;
	Int         fi, fStart, fEnd, fLen;
	SLContext	*context = dec.context;
	PointList	*points = SC_GET(Points);
	IndexList	*indexes = SC_GET(PointIndexes);
	IndexList   *faces = SC_GET(FaceIndexes);

	if (!points)
		return;

	if (!indexes)
	{
		if (!gIdentityList)
		{
			gIdentityList = new IndexList(256);
			for (i = 0; i < gIdentityList->NumItems(); i++)
				gIdentityList->Item(i) = i;
		}
		indexes = gIdentityList;
		Assert(points->NumItems() <= 256, "polygon too large");
		// note: we are relying on Array not changing the current items
		// array when we resize down and then back up again. 
		// (This is currently safe: the items array only changes if the array
		// size is increased.)
		indexes->SetSize(points->NumItems());
	}

	changed = 1;
	if (faces)
	{
		fStart = (*faces)[0];
		fEnd = (*faces)[1];
		fLen = fEnd - fStart;
		fi = 2;
	}
	else
	{
		fStart = 0;
		fLen = fEnd = indexes->NumItems();
	}

	while (true)
	{
		if (!(dec.flags & DecTris))
			n = fLen;
		else if (fLen == 4 && (dec.flags & DecQuads))
		{
			// XXX : not stable as height of pgram goes to zero.
			
			Int		*ip = &(*indexes)[fStart];
			Vector	ab = (*points)[ip[1]] - (*points)[ip[0]];
			Vector	cd = (*points)[ip[2]] - (*points)[ip[3]];

			// ab == cd : |ab - cd| / |ab + cd| < eps?

			if (sqrlen(ab - cd) / sqrlen(ab + cd) < 1e-6)
				n = 4;
			else
				n = 3;
		}
		else
			n = 3;

		for (j = 0; j < n; j++)
			idx[j] = (*indexes)[j + fStart] + dec.pointsAccNum;
						
		dec.HandlePoly(n, idx, changed);
		dec.decNum++;
		changed = 0;
		
		if (n == 3)
			for (j = fStart + 2; j < fEnd - 1; j++)
			{
				idx[1] = (*indexes)[j] + dec.pointsAccNum;
				idx[2] = (*indexes)[j + 1] + dec.pointsAccNum;
				dec.HandlePoly(n, idx, 0);
				dec.decNum++;
			}

		if (faces && fi < faces->NumItems())
		{
			fStart = fEnd;
			fEnd = (*faces)[fi++];
			fLen = fEnd - fStart;
		}
		else
			break;
	}
}

Void scPoly::Parse(istream &s)
{
}


// --- Attributes -------------------------------------------------------------
    
Void scPoints::Apply(const Transform &m)
{
	Int i;
	
	for (i = 0; i < NumItems(); i++)
		SELF[i] = proj(xform(m, Vector4(SELF[i], 1.0)));
}

Void scCoords::Apply(const Transform &m)
{
}

Void scTransform::ApplyActionSelf(scSceneAction &a)
{
	scObject::ApplyActionSelf(a);
	a.transAcc = xform(a.transAcc, SELF);
}

Void scCamera::Parse(istream &s)
{
	String	str;
	Vector	p;
	Char	c;
	Int		model = 1;

	ChompWhiteSpace(s);
	while (s.peek() == '+')					// extension?
	{
		s.get(c);
		str.ReadWord(s);

		if (str == "eye")
			s >> position;
		else if (str == "dir")
			s >> viewDir;
		else if (str == "lookAt")
		{
			s >> lookAt;
			model = 2;
		}
		else if (str == "vup")
			s >> up;
		else if (str == "scale")
			s >> scale;
		else if (str == "fov")
			s >> fov;
		else if (str == "clipNear")
			s >> clipNear;
		else if (str == "clipFar")
			s >> clipFar;
		else
			cerr << "unknown camera option: '" << str << "'" << endl;
		
		ChompWhiteSpace(s);
	}

	SetupFromParams();
}

Void scCamera::Print(ostream &s) const
{
//	GetViewDirParams(); 	XXX can't do this 'cause of const
	
	s << "camera";
	
	if (scale != 1.0)
		s << " +scale " << scale;

	s << " +eye " << position;
	s << " +dir " << viewDir;
	s << " +vup " << up;
	s << " +fov " << fov;
}

Void scIndexes::Print(ostream &s) const
{
	switch (AttrID())
	{
	case aPointIndexes:
		s << "indexes ";
		break;
	case aNormalIndexes:
		s << "n_indexes ";
		break;
	case aCoordIndexes:
		s << "t_indexes ";
		break;
	case aColourIndexes:
		s << "c_indexes ";
		break;
	case aFaceIndexes:
		s << "faces ";
		break;
	}
	
	s << (IndexList&) SELF << endl;
}

Void scTexture::Parse(istream &s)
{
	String str;
	
	ChompWhiteSpace(s);
	s >> str;
	Load(str);
}

Void scTexture::Print(ostream &s) const
{
	s << "texture \"" << textureFile.GetPath() << '"';
}

Void scMeshType::Parse(istream &s)
{
	String str;
	
	ChompWhiteSpace(s);
//	s >> itsType;
}

Void scMeshType::Print(ostream &s) const
{
	s << "meshtype \"" << itsType << '"';
}

