/*
	File:			HProgRad.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1999-2000, Andrew Willmott

	Notes:			

	Change History:
*/

#include "HProgRad.h"
#include "HierMesh.h"
#include "gcl/Draw.h"


Void HProgRad::ResetOptions()
{
	Int		i;
	
	CreatePatches();

	elements.Clear();
	eltParents.Clear();

	for (i = 0; i < patches.NumItems(); i++)
		((HierElem *) patches[i])->CreateElements(elements, eltParents, i);  

	ColourMeshInitial();
	ColourVertices();

#ifdef RAD_VIS
	Field(out1) << patches.NumItems() << " patches, " << 
		elements.NumItems() << " elements." << show;
#endif
}

static Colour SingleFormFactor(RadElem *fromPatch, RadElem *toPatch)
{
	GCLReal 	visibility;
	GCLReal		factor;
	Colour		result;
		
	if (len(toPatch->Reflectance()) < 0.0001)
	{
		result.MakeZero();
		return(result);
	}

	visibility = fromPatch->Visibility(toPatch);
		
	if (visibility > 0.0)
	{
		factor = fromPatch->EstFormFactor(toPatch);
		result = toPatch->Reflectance();
		result *= factor * visibility;
	}
	else
		result.MakeZero();
		
	return(result);
}


Bool HProgRad::Render()
{
	Int			i, j;
	Colour		c;
	Colour		deltaRad;
	Colour		radToShoot;
	Bool		subdivided;
	Int			numElts = elements.NumItems();
	Int 		numPatches = patches.NumItems();
	GCLReal		power, envArea;
	Colour		envPower, envRefl;
	Bool		finished = false; 
	RadElem		*shootPatch;
	ColourList	*theFFRow;
	
	RadMethod::Render();
	
	iterations = 0;
	error = 1.0;
	origShoot = 0;
	
	B.SetSize(numElts);			// Radiosity vectors for R, G, B
	S.SetSize(numPatches);		// Unshot radiosity vectors for R, G, B
	
	for (i = 0; i < patches.NumItems(); i++)
		S[i] = patches[i]->Emittance();
	
	for (i = 0; i < numElts; i++)
		B[i] = S[eltParents[i]];

	envRad = vl_1;

	if (Stage(1)) return(0);
	
	//	Setup for the ambient term

	envRefl.MakeZero();
	envArea = 0;
	envPower.MakeZero();
	 
	for (i = 0; i < numPatches; i++)
	{
		envRefl += patches[i]->area * patches[i]->Reflectance();
		envArea += patches[i]->area;	
		envPower += S[i] * patches[i]->area;
	}
	
	envPower /= envArea;

	//	R = 1 / (1 - avgReflectance)
	
	for (i = 0; i < 3; i++)
		envRefl[i] = 1.0 / (1.0 - envRefl[i] / envArea);
				
	//	Now, run the algorithm...
						
	while (!finished)
	{	
		if (Stage(2)) return(0);
		
		iterations++;
				
		//	Find the patch with the max power to shoot...
				
		maxPower = 0;
		maxPowerIndex = 0;
		envPower.MakeZero();
		
		for (i = 0; i < numPatches; i++)						
		{
			c = S[i] * patches[i]->area;			
			power = dot(c, kRadRGBToLum);
			
			if (maxPower < power)
			{
				maxPower = power;
				maxPowerIndex = i;
			}
		}
		
		radToShoot = S[maxPowerIndex];
		radToShoot *= gRadControl->alpha;
		shootPatch = patches[maxPowerIndex];

		if (Stage(3)) return(0);

		// Find initial form factors
#ifdef RAD_VIS
		if (gRadControl->drawMatrix)
		{
			if (maxPowerIndex >= FFElts.NumItems())
				FFElts.SetSize(maxPowerIndex + 1);
	   		FFElts[maxPowerIndex] = elements;

			theFFRow = &FFRows[maxPowerIndex];
		}
		else
#endif
		theFFRow = &FFRow;
		MakeFormFactorFromVector(*shootPatch, elements, *theFFRow);

		if (Stage(4)) return(0);			
	
		//	Shoot radiosity, updating S and B as necessary			
		error = 0;
		S[maxPowerIndex] -= radToShoot;
			
		for (i = 0; i < numElts; i++)
		{
			deltaRad = (*theFFRow)[i] * radToShoot;
			B[i] += deltaRad;
										
			// Push radiosity delta to the parent...					
			S[eltParents[i]] += deltaRad * (elements[i]->area 
											/ patches[eltParents[i]]->area);
			error += sqrlen(deltaRad);
		}
	
		error = sqrt(error);

		// Calculate ambient light in scene...
	
		if (gRadControl->ambient)
		{
			envPower = vl_0;
			for (i = 0; i < numPatches; i++)
				envPower += S[i] * patches[i]->area;
			
			envPower /= envArea;
			envRad = envRefl * envPower;
		}
		else
			envRad.MakeZero();

		if (true)	// was multigrid option???
			do 
			{						
				Colour		oldFF, f0, f1, f2, f3, temp;

				// Subdivide receiving elements as necessary...
				
				subdivided = false;
				
				// Distribute radiosity to vertices so we can measure error
				for (i = 0; i < numElts; i++)
					((HierElem *) elements[i])->SetColour(
						B[i] + envRad * elements[i]->Reflectance());
				ColourVertices();
				
				if (Stage(5)) return(0);

				for (i = 0; i < numElts; i++)
				{
					HierElem *old = (HierElem *) elements[i];
				
					if (old->area >= gRadControl->kAError &&
							old->RadError() > gRadControl->kFError)
					{
						// We need to subdivide
												
						if (!old->HasChildren())
							old->Subdivide();

						// Replace elt with its subdivisions...
						elements[i] = old->child[0];	   						
						elements.Append(old->child[1]);
						elements.Append(old->child[2]);
						elements.Append(old->child[3]);
						
						j = eltParents[i];
						eltParents.Append(j);
						eltParents.Append(j);
						eltParents.Append(j);
																		
						// Generate new form factors to the children...
						f0 = SingleFormFactor(shootPatch, old->child[0]);
						f1 = SingleFormFactor(shootPatch, old->child[1]);
						f2 = SingleFormFactor(shootPatch, old->child[2]);
						f3 = SingleFormFactor(shootPatch, old->child[3]);
						oldFF = (*theFFRow)[i];
						
						theFFRow->Item(i) = f0;
						theFFRow->Append(f1);
						theFFRow->Append(f2);
						theFFRow->Append(f3);
						
						// update contribution to parent...		
						temp = f0 * old->child[0]->area 
							+ f1 * old->child[1]->area 
							+ f2 * old->child[2]->area 
							+ f3 * old->child[3]->area 
							- oldFF * old->area;
						S[j] += radToShoot * (temp / patches[j]->area);

						// update child radiosities...					
						B.Append(B[i] + radToShoot * (f1 - oldFF));
						B.Append(B[i] + radToShoot * (f2 - oldFF));
						B.Append(B[i] + radToShoot * (f3 - oldFF));
						B[i] += radToShoot * (f0 - oldFF);

						subdivided = true;
					}	
				}
				numElts = elements.NumItems();
#ifdef RAD_VIS
				if (maxPowerIndex >= FFElts.NumItems())
					FFElts.SetSize(maxPowerIndex + 1);
				FFElts[maxPowerIndex] = elements;
#endif											
				
				Stage(6);
			} 
			while (subdivided);
	
		// Stop us before we shoot again

		if (iterations == 1)
			origShoot = maxPower;
		else
			finished = 
				(gRadControl->maxShots > 0 && iterations > gRadControl->maxShots)
				|| (maxPower < gRadControl->error * origShoot);
		
		if (Stage(7)) return(0);
	}	
	
	if (Stage(8)) return(0);

	return(1);
}


#ifdef RAD_VIS

Void HProgRad::DumpStats(){}

Int HProgRad::Stage(Int stage)
{
	Int		i, j;
	Bool	animate = gRadControl->animate || gRadControl->pause;
 
	switch (stage)
	{
	case 1:		// pre setup
		Field(out3) << patches.NumItems() << " patches, " 
					<< elements.NumItems() << " elements." << show;
		Field(out1) << "Forming emission vector E..." << show;
		FFRow.Clear();
		FFElts.Clear();
		if (gRadControl->drawMatrix)
		{
			FFRows.SetSize(patches.NumItems());
			FFElts.SetSize(patches.NumItems());
//			FFRows.ClearTo(ColourList());
//			FFElts.ClearTo(PatchList());
		}
		StartUpdate();
		break;
		
	case 2:		// post setup
		ColourVertices();
		display->Redraw();

		doUpdate = animate || Update();

		if (animate)
			Field(out1) << "Finding patch to shoot..." << show;
		else if (doUpdate)
		{
			Char *plural;
			if (iterations == 1)
				plural = "";
			else
				plural = "es";
			Field(out1) << iterations << " patch" << plural << " shot" << show;
			
		}			
		break;
		
	case 3:		
		if (doUpdate)
		{
			if (animate)
			{
				Field(out1) << "Finding factors for patch " << maxPowerIndex 
							<< show;
				
			// flash shooting patch red
				patches[maxPowerIndex]->SetHighlight(1);
			}

			ColourVertices();
			display->Redraw();		
		}
		break;
		
	case 4:
		if (doUpdate)
		{
			if (animate)
				Field(out1) << "Shooting patch " << maxPowerIndex << show;
			RenderMatrix();
			display->Redraw();		
		}
		break;
		
	case 5:			
		if (animate)
			Field(out1) << "Looking for elements to subdivide..." << show;
		if (doUpdate)
			display->Redraw();
		break;
		
	case 6:	
		if (doUpdate)
		{
			display->Redraw();
			UpdateCont();
		}
		if (animate)
			Field(out1) << "Re-shooting Patch " << maxPowerIndex << show;
		Field(out2) << "error = " << maxPower / origShoot 
					<< ", RMS Error = " << error << show;
		Field(out3) << patches.NumItems() << " patches, " 
					<< elements.NumItems() << " elements." << show;
		break;
		
	case 7:		// Middle of solve loop.
		if (doUpdate)
		{
			Colour c;
		
			patches[maxPowerIndex]->SetHighlight(0);

			if (gRadControl->shotDisplay)
				// Colour the elements according to S...
				for (i = 0; i < elements.NumItems(); i++)
					((HierElem*) elements[i])->SetColour(S[eltParents[i]]);
			else
				// Colour the elements according to B...
				for (i = 0; i < elements.NumItems(); i++)
					((HierElem*) elements[i])->
						SetColour(B[i] + envRad * elements[i]->Reflectance());
		
			ColourVertices();
			display->Redraw();
		
			if (animate)
			{
				Field(out1) << "Done with shooting " << show;
				Field(out3) << patches.NumItems() << " patches, " 
							<< elements.NumItems() << " elements." << show;
		
				Field(out2) << "error = " << maxPower / origShoot 
							<< ", RMS Error = " << error << show;
			}
			UpdateCont();
		}
		break;
		
	case 8:
		Field(out1) << "Finished: " << iterations << " patches shot" << show;
		
		// Colour the elements according to B...
		for (i = 0; i < elements.NumItems(); i++)
			((HierElem*) elements[i])->SetColour(B[i] + envRad 
												 * elements[i]->Reflectance());
		ColourVertices();
		break;
	}
	
	return(Pause());
}

#else

Void HProgRad::DumpStats()
{
	Int		i, j;
	Colour	c;
	GCLReal	shotErr, mem;

	if (origShoot == 0.0)
		shotErr = 1.0;
	else
		shotErr = maxPower / origShoot;
	
	mem = sizeof(Colour) * (B.NumItems() + S.NumItems() + FFRow.NumItems()
			  + elements.NumItems()) + sizeof(Int) * eltParents.NumItems();
	mem /= 1024.0;
	
	cout << dumpID
		<< ' ' << totTime
		<< ' ' << gRadControl->stage
		<< ' ' << patches.NumItems()
		<< ' ' << elements.NumItems()

		<< ' ' << shotErr
		<< ' ' << error
		<< ' ' << gRadControl->rays
		<< ' ' << iterations
		<< ' ' << mem
		<< ' ' << grid->MemoryUsage()
 		<< ' ' << TotalMemoryUse()
		<< endl;

	for (i = 0; i < elements.NumItems(); i++)		// Colour the elements...
		((HierElem*) elements[i])->SetColour(B[i] + 
							  envRad * elements[i]->Reflectance());

	DumpScene();
#ifdef RAD_MEM_USE
	DumpMemoryUse();
#endif
}

Int HProgRad::Stage(Int stage)
{
	if (CheckTime()) return(1);

	gRadControl->stage = stage;

	switch (stage)
	{
	case 1:		// pre setup
		cout << "method hprog " << endl;
		cout << "sub " << gRadControl->patchSubdivs << endl;
		cout << "esub " << gRadControl->eltSubdivs << endl;
		cout << "tErr " << gRadControl->error << endl;
		cout << "fErr " << gRadControl->kFError << endl;
		cout << "aErr " << gRadControl->kAError << endl;
		cout << "scene " << sceneName << endl;
		cout << "polys " << gRadControl->numPolys << endl;
		cout << "srcPatches " << patches.NumItems() << endl;
		cout << "format "
			 << "ID "
			 << "time "
			 << "stage "
			 << "srcPatches "
			 << "patches "
			 << "shootErr "
			 << "resErr "
			 << "rays "
			 << "iterations "
			 << "mem "
			 << "rtMem "
			 << "rss "
			 << endl;
 
		cout << "----------------------------------------------"
			 << "---------------------------------" << endl;
	
		DumpStats();
		break;
				
	case 8:
		DumpStats();
		break;
	}
	
	if (Idle()) return(1);
	timer.ContTimer();
	return(0);
}

#endif

Void HProgRad::DrawMatrix(Renderer &r)
{
	if (!gRadControl->drawMatrix || gRadControl->stop)
		return;
		
#ifdef RAD_VIS
	Int			i, j;
	GCLReal		x1, x2, y1, y2, wx, wy, m;
	HierElem	*hq;
	
	wx = 2.0 / FFRows.NumItems();
	
	x1 = -1;
	x2 = wx - 1;
	
	for (i = 0; i < FFRows.NumItems(); i++)
	{
		y2 = 1;
		
		// Draw a column
		
		for (j = 0; j < FFRows[i].NumItems(); j++)
		{
			hq = (HierElem*) FFElts[i][j];
			y2 = 1 - 2 * GCLReal(eltParents[j]) / GCLReal(patches.NumItems());

			wy = 2.0 / GCLReal(patches.NumItems());
			m = GCLReal(1 << (2 * hq->level));
			wy /= m;
	
			y2 -= GCLReal(hq->treeCode) * wy;	
			y1 = y2 - wy;
		
			if (hq->highlight == 2 && patches[i]->highlight == 3)
				r.C(cPurple);
			else if (hq->highlight == 2)
				r.C(cYellow);
			else if (patches[i]->highlight == 3)
				r.C(cGreen);
			else	
				r.C(FFRows.NumItems() * FFRows[i][j]);

			PaintRect(r, Coord(x1, y1), Coord(x2, y2)); 

			y2 = y1;
		}
		
		x1 += wx;
		x2 += wx;
	}
#endif
}

RadElem *HProgRad::NewMesh()
{
	if (gRadControl->noGridMesh)
		return(new HierElem());
	else
		return(new HierGrid());
}

Void HProgRad::ColourVertices()
{
	Int		i;

	for (i = 0; i < elements.NumItems(); i++)
		((HierElem*) elements[i])->CorrectLeaves();

	RadMethod::ColourVertices();
}

PatchList *HProgRad::GetElements()
{
	Int 			i, j, n, numElems, numSplitElems = 0;
	HierElemPtr		subElems[5];
	
	// Make the mesh compliant, if it's been tesselated.

	ColourVertices();

	if (gRadControl->anchor)
	{
		n = elements.NumItems();
		for (i = 0; i < n; i++)
		{
			((HierElemPtr) elements[i])->SplitElement(numElems, subElems);

			if (numElems > 0)
			{
				numSplitElems++;

				
				elements[i] = subElems[0];
				for (j = 1; j < numElems; j++)
					elements.Append(subElems[j]);
			}
		}

	}

	return(&elements);
}

Void HProgRad::DumpMemoryUse()
{
	Float	propmem, bmem, bpmem, pmem, emem, semem, epmem, total;
	Float	ptmem, clrmem, rtmem;
	Int		i;
	Char	*fmt = "%+10s %10.1f\n";
	
	propmem = sizeof(RadProps) * numProps / 1024.0;
	ptmem = sizeof(Point) * points->NumItems() / 1024.0;
	clrmem = sizeof(Colour) * colours->NumItems() / 1024.0;

	bmem = sizeof(RadElem*) * baseElems.NumItems() / 1024.0;
	bpmem = 0;
	RadElem::sGridMem = 0;
	RadElem::sGridChildMem = 0;
	for (i = 0; i < baseElems.NumItems(); i++)
		bpmem += baseElems[i]->MemoryUse();
	bpmem /= 1024.0;
	
	pmem = sizeof(RadElem*) * patches.NumItems() / 1024.0;
	emem = sizeof(RadElem*) * elements.NumItems() / 1024.0;
	epmem = sizeof(Int) * eltParents.NumItems() / 1024.0;
	rtmem = grid->MemoryUsage();
	
	printf(fmt, "propmem", propmem);
	printf(fmt, "points", ptmem);
	printf(fmt, "colours", clrmem);
	printf(fmt, "bmem", bmem);
	printf(fmt, "bpmem", bpmem);
	printf(fmt, "  grid", RadElem::sGridMem / 1024.0);
	printf(fmt, "  gridCh", RadElem::sGridChildMem / 1024.0);
	printf(fmt, "pmem", pmem);
	printf(fmt, "emem", emem);
	printf(fmt, "semem", semem);
	printf(fmt, "epmem", epmem);
	printf(fmt, "rtmem", rtmem);

	total = propmem + bmem + bpmem + pmem + emem + semem + epmem
		 + clrmem + ptmem + rtmem;

	printf(fmt, "total", total);
	grid->DumpMemoryUsage();
}
