/*
	File:		MeshJoin.h
	
	Function:	Connect up adjacent triangles in a shared-vertex mesh.
				Sets the 'nbFace' and 'nbEdge' fields correctly for
				groups of triangles belonging to the same surface.

				This representation consists of
					+ faces with a fixed number of edges
					+ neighbour edge pointers
						for each edge of the face, you can find the corresponding
						edge of the neighbouring face.

				This is not as general as a full winged-edge structure, but
				more space efficient in a mesh that will only have triangles
				(or quads in our case). Drawbacks:
					+ can't handle input meshes with t-intersections
					+ can't get from vertex to neighbouring faces
						can solve by keeping 1:1 vertex->face list
	
	Notes:		Requires triangles to share a single vertex array
				
				Calling conventions: 
					Call Init() with the number of vertices in the model
					Call StartGroup() for each surface group
					Call AddTriangle for all faces in that group
					Call EndGroup() when done.
					Repeat the last 3 calls as necessary,  then call Finished().

	Author:		Andrew Willmott, 1997
*/

#ifndef __MeshJoin__
#define __MeshJoin__

/*
	Data structures are face-orientated: each triangle
	contains three neighbour fields, which point to
	the neighbouring triangle along each edge, as well
	as an edgeIndex, which tells us which edge of the
	neighbour triangle abuts ours.

	One of these (face, edgeIndex) pairs defines
	an edge. As in, edge 'edgeIndex' of face 'face'.
*/

// For HierElems ...

#include "RadMesh.h"

typedef IndexList			EdgeList;
typedef Byte	   			Edge;
typedef NbRadElem			Face;
typedef Face				*FacePtr;

// the rest of the code is more general, though EdgesMatch() will
// require customisation.

class MeshJoin
{
public:
	Void		Init(Int numVertices);
	Void		StartGroup();
	Void		AddTriangle(Face *addFace);
	Void		EndGroup();
	Int			Finished();
	
protected:
	Bool		EdgesMatch(
						FacePtr face1,
						Edge	edge1,
						FacePtr face2,
						Edge	edge2
					);
	Void		SetFaceChainVertices(Face *f, Int edge, Int i);

	FacePtr		*vertexFaces;
	Edge		*vertexEdges;
	Int			numVertices;
	Int			surfaceVertices;
	
	// together these define the start of a linked list of edges
	// for each vertex.
};

#endif

