/*
	File:			RadControl.h

	Function:		Defines a structure for all of the possible options for
					the various radiosity illumination methods.
					
	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
 */

#ifndef __RadOptions__
#define __RadOptions__

#include "gcl/Renderer.h"
#include "gcl/Scene.h"

enum RadType
{
	kNoMethod,
	kMatrix,
	kProgressive,
	kProgSubstruct,
	kHierarchical, 
	kAnalytical,
	kNumRadMethods
};

enum HRBasisType
{
	kHaar = 1,
	kFlatlet2,
	kFlatlet3,
	kMultiwavelet2,
	kMultiwavelet3
};

/// inter-patch visibility methods
enum VisType	
{
	vis_none,	///< don't check visibility
	vis_1,		///< cast a single ray between the centre of each patch
	vis_16x1,   /**< cast 16 rays from centre of the receiving patch to 
					 the source patch. */
	vis_4x4		/**< cast 16 rays from anywhere on source to 16 anywhere on
					 dest. Uses Cohen's jittered magic-square method. */
};

enum MeshType
{
	mesh_std,		///< standard, divide evenly by area 
	mesh_random,	///< vary mesh density for different patches
	mesh_nonlin		///< mesh more heavily at patch edges.
};

enum SolverType
{
	sv_twoStage,	///< Refine, then solve///< standard, divide evenly by area 
	sv_interleaved,	///< Intermix refinement & solution
	sv_scheduled	///< Repeat refine/soln. for progressively smaller eps.
};

class HRElem;
class RadMethod;

/** A 'fat' structure containing all the options any of the radiosity
	methods may need. */
class RadControl			
{							
	public:					

	RadControl();

	//	Options for the current render

   	RadType			method;
   	HRBasisType		basis;

	//	Method options

	GCLReal		patchSubdivs;	///< How fine to mesh the scene
	GCLReal 	eltSubdivs;		///< How fine to mesh sub elements
	GCLReal 	alpha;			///< For overrelaxation solution
	GCLReal 	error;			///< error level at which to terminate
	
   	GCLReal		kFError;		///< Allowable error in form factors
   	GCLReal		kAError;		///< Minimum element size
	SolverType	solver;			///< HR solution algorithm to use
   	Bool 		useBF;			///< Do brightness-weighted refinement
   	Bool 		cluster;		///< Use clustering
   	Bool 		gridOn;			///< Ray-trace grid on
   	Bool 		textureRefl;	///< Incorporate texture info into reflectance calcs.
   	Bool 		useConjGrad;	///< Use conjugate gradient solver (matrix)
   	Bool 		ambient;		///< Add in ambient-light correction.
	Bool		refAllLinks;	/**< ensure all links have been refined before
									 terminating. */
	Int			refineLevels;	/**< maximum number of levels to refine a link
									 on each iteration. */
	Int			maxShots; 		///< Max number of shooting steps.
	Int			schedIterations; ///< How many sched. solver iterations to perform.

	//	Visibility and form factor stuff.

   	VisType		visibility;		///< Type of visibility testing to use
	Bool		jitterRot;		/**< Recalculate jitter for each new visibility
									 test. */
	GCLReal		dFError;		/**< Error tolerance for the differential FF
									 estimate before swapping to analytical FF. */
	GCLReal		visError;		/**< How much more to subdivide when visibility 
									 is partial in wavelet radiosity: in such 
									 situations the FF limit becomes kFEError *
									 visError. */
	Int			quadLevel;		///< How accurate to make quadrature
	GCLReal		forceVisReuseArea;	/**< links to patches greater than this size
									always perform visibility calcs. */
	GCLReal		sampleFactor;	///< how many samples to use for final gather...
	Bool		visInQuad;		/**< use fractional visibility (a la original WR
									 paper) or integrate into quadrature. */
	GCLReal		favourReceiver; /**< > 1.0 means more likely to subdivide 
									 receiver. */

	//	Mesh options

	MeshType	mesh;			///< type of meshing to employ.
	Bool		fixMesh;		///< attempt to share mesh vertices.
	Bool		connectMesh;	///< calculate adjcency info for base mesh.
	Bool		noGridMesh;		///< don't use gridding; hierarchical refinement only

	// MR complexity options

	GCLReal		rtComplexity;	///< model complexity to use for raytracing
	GCLReal		meshComplexity; ///< model complexity to use for radiosity mesh.

	//	Simulation opts.

	Bool 		stop;
	Bool 		step;
	Bool		pause;

	//	Display options
		
	Bool		graded;			///< Adjacent leaves not more than one level apart
	Bool		anchor;			///< Eliminate T-vertices before display
   	Bool		gouraud;		///< Gouraud-shade patches
   	Bool		texture;		///< Enable texture-mapping
   	Bool 		wire;			///< Wireframe.

   	Bool		redWire;		///< Wireframe, but colour red.
   	Bool		shotDisplay;	///< show unshot rad. rather than accumulated 
   	Bool		patchView;		///< Display relevant to currently selected patch
   	HRElem		*pvData;		///< Hook for the above
   	Bool		funcView;		///< Plot rad as function over patches
   	
   	Bool		animate;		///< Animate by highlighting relevant patches
   	Bool		showRays;		///< Animate casting of rays
   	Bool		showLinks;      ///< show links being created
   	Bool		showFCs;        ///< show face clusters
   	
   	Int			choke;			/**< if non-zero, only display hierarchy down
									 to given level. */
	Bool		drawMatrix;		///< Whether or not to display matrix.
	
	Bool		outlineClusters;
	Bool		outlineVisGrid;

	Bool		finalPass;		///< doing final illuminate pass
	Bool		bestPass;		///< refine mesh further on final pass
	Bool		bestVisPass;	///< high quality vis on final pass
	Int			bestLevels;		///< extra levels of subdivision to employ

	// debugging, adding code

	Bool		o1, o2, o3, o4, o5, o6;		///< debugging etc.
	
	//	Objects relevant to the current render
	   	
	RadMethod	*radObject;
	Int			stage;
	
	//	Statistics for the current render
	
	Bool		dumpTree;
				///< dump text version of radiosity hierarchy
	Bool		dumpScenes;
				///< dump scene when dumping stats.
	Bool		updateScene;
				///< rewrite output mesh on each time slice
	GCLReal		sliceTime, limitTime;
   	Int			numPolys;
	Int			rays;
	GCLReal		mem;
	StrConst	outFile;
};

extern Reflectance	kRadRGBToLum;
extern Int 			radVersionMajor;
extern Int 			radVersionMinor;
extern Int 			radVersionFix;
extern RadControl	*gRadControl;

TempString 			RadGetVersion();
StrConst 			RadGetCopyright();

#endif
