/*
	File:			RadMain.cc

	Function:		Implements "Radiator"
	
					Provides command line-driven program for running 
					various radiosity methods. 

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

	Notes:			
*/

#ifdef CL_HAS_NEW
#include <new>
#endif
#include <unistd.h>
#include <fstream.h>
#include <signal.h>

#include "cl/ArgParse.h"

#include "RadMethod.h"

#include "cl/String.h"
#include "gcl/Avars.h"
#include "gcl/SceneLang.h"
#include "gcl/Readers.h"

#include "BuildDate.h"

class RadMain
{
public:
	Void 		SetOptions(Int argc, Char *argv[]);
	Void		DoRadiosity();

	RadControl	radOpts;
	Int			size;
	Int			dumpObj;
	Int			dumpSL;
	Int			noDirect;
	Int			normalise;
	Char		*sceneFile;
	String		outFile;

	static Int	frame;
};

static char *defMethodName[] = 
{
	"none",
	"mat",
	"prog",
	"hprog",
	"hier",
	"ana"
};

static char *defBasisName[] = 
{	
	"none",
	"haar",
	"f2",
	"f3",
	"m2",
	"m3"
};


// --- Utilities --------------------------------------------------------------

static void LogToFile(const FileName &logFile)
// capture stdout & copy it to a run file
{
	int		filedes[2], count;
	char	buffer[1024];
	FILE 	*log;

	if (pipe(filedes) == -1)
	{
		perror("logger: pipe");
		return;
	}
	
	fflush(stdout);

	if (fork() != 0)
	{
		if (dup2(filedes[1], fileno(stdout)) == -1)
			perror("logger: dup");
		if (close(filedes[0]) == -1)
			perror("logger: c0");

		return;
	}
	
	close(filedes[1]);
	log = logFile.FOpen("w");

	while ((count = read(filedes[0], buffer, 1024)) > 0)
	{
		write(1, buffer, count);
		fwrite(buffer, 1, count, log);
	}
	if (count == -1)
		perror("logger: read");

	fclose(log);
	close(filedes[0]);

	exit(0);
}

static Void RadHandleInterrupt(Int a)
{
// On reception of a SIGHUP, stop simulation.
	cerr << "received interrupt " << a << ": stopping simulation" << endl;
	gRadControl->limitTime = 0;
}

static Void RadHandleCont(Int a)
{
// On reception of a SIGCONT, print status info
	cerr << "received interrupt " << a << ": printing status" << endl;
	gRadControl->radObject->lastTime = gRadControl->radObject->totTime
										- gRadControl->sliceTime;
}

static Array<Char*> gAvarNames;
static Array<GCLReal> gAvarVals;

static Int avarArg(int argc, char **argv)
{
	Int i;
	GCLReal val;
	
	if (argc % 2 == 1)
	{
		cerr << "Wrong number of avar arguments." << endl;
		exit(1);
	}
	
	for (i = 0; i < argc; i += 2)
	{
		val = atof(argv[i + 1]);
		if (argv[i] == StrConst("frame"))
			RadMain::frame = (Int) val;
		gAvarNames.Append(argv[i]);
		gAvarVals.Append(val);
	}
	
	return(0);
}


// --- RadMain methods --------------------------------------------------------

Int	RadMain::frame = -1;

Void RadMain::SetOptions(Int argc, Char *argv[])
{
	Int 		hier, prog, mat, hprog, ana;
	Int 		haar, f2, f3, m2, m3, noClusFlag;
	Int			schedFlag, interFlag, twoStageFlag, fcrFlag = false;
	Int 		ambient, viq, bestPass, bestVisPass;
	Int 		noBF, noGraded, noAnchor, refAllLinks, conjGrad;
	Int			visnone, vis1, vis16, vis44;
	Int			updateScene, dumpMatrix, dumpScenes, dumpTree;
	Int			meshRnd, meshNL, meshFix, meshNoConnect, meshNoGrid;
	Int			version;
	Double		meshComp, rtComp;
	String		usage;
	Arg_form	*arg_format;
	
	// Set default arguments
	
	size = 400;
	frame = -1;
	sceneFile = 0;
	radOpts.outFile = 0;
	meshComp = 1.0;
	rtComp = -1.0;

	usage.Printf("Usage: %s [options], where options are as follows:", 
			argv[0]);

	// Create arg form
	
	arg_format = arg_to_form(0,
		"-v", ARG_FLAG(&version), 					"show current version",
		"[%S]",	&sceneFile,							"input scene file",

// methods

		"", 										
		"\n--- Methods --------------------------------------------------------------------\n",
		"-mat", ARG_FLAG(&mat), 					"use matrix radiosity",
		"-prog", ARG_FLAG(&prog), 					"use progressive radiosity",
		"-hprog", ARG_FLAG(&hprog), 				"use progressive radiosity with substructuring",
		"-hier", ARG_FLAG(&hier), 					"use hierarchical radiosity",
		"-ana", ARG_FLAG(&ana), 					"use 1st-order analytical solver",
		
// method params
		
		"", 										
		"\n--- Params ---------------------------------------------------------------------\n",
		"-sub %F", &radOpts.patchSubdivs, 			"set patch subdivision density",
		"-esub %F", &radOpts.eltSubdivs, 			"set element subdivision density",
		"-alpha %F", &radOpts.alpha, 				"set alpha",
		"-error %F", &radOpts.error, 				"set termination error",
		"-ferr %F", &radOpts.kFError, 				"set ff error",
		"-aerr %F", &radOpts.kAError, 				"set min patch size",
		"-derr %F", &radOpts.dFError, 				"set singularity threshold",

// bases


// visibility, form factor eval

		"", 										
		"\n--- Visibility------------------------------------------------------------------\n",
		"-verr %F", &radOpts.visError, 				"wavelet visibility error",
		"-quadLevel %d", &radOpts.quadLevel, 		"quadrature level",
		"-visInQuad", ARG_FLAG(&viq), 				"do visibility testing in quadrature",
					
		"-vis_none", ARG_FLAG(&visnone), 			"no visibility testing",
		"-vis_1", ARG_FLAG(&vis1), 					"single ray visibility", 
		"-vis_16", ARG_FLAG(&vis16), 				"use 16 rays from source -> centre of dest",
		"-vis_44", ARG_FLAG(&vis44), 				"use 16 rays distributed over both patches",
		"-visReuseArea %F", &radOpts.forceVisReuseArea, 
													"area below which to reuse visibility samples",
		"-favourReceiver %F", &radOpts.favourReceiver,
													"amount by which to favour subdividing receiver",

// meshing

		"", 										
		"\n--- Meshing --------------------------------------------------------------------\n",
		"-mesh_share", ARG_FLAG(&meshFix), 			"ensure all coincident mesh vertices are shared",
		"-mesh_noAnchor", ARG_FLAG(&noAnchor), 		"don't enforce elimination of t-vertices",
		"-mesh_noGrading", ARG_FLAG(&noGraded), 	"don't enforce a graded mesh",
		"-mesh_noConnect", ARG_FLAG(&meshNoConnect), "don't calculate mesh connectivity",
		"-mesh_noGrid", ARG_FLAG(&meshNoGrid), 		"don't use gridding, only hierarchical refinement",
		"-mesh_nonLin", ARG_FLAG(&meshNL), 			"mesh more heavily at edges",
		"-mesh_random", ARG_FLAG(&meshRnd), 		"vary 'sub' parameter randomly",

// MRM complexity

		"", 										
		"\n--- Complexity -----------------------------------------------------------------\n",
		"-meshComp %F", &meshComp,				 	"complexity for creating radiosity mesh",
		"-rtComp %F", &rtComp,						"complexity for raytracing (if different.)",

// misc

		"", 										
		"\n--- Misc -----------------------------------------------------------------------\n",
		"-conjGrad", ARG_FLAG(&conjGrad), 			"use conjugate gradient solver (matrix radiosity)",
		"-ambient", ARG_FLAG(&ambient), 			"add ambient term",
		"-maxShots %d", &radOpts.maxShots, 			"max shooting steps to perform",
		"-noDirect", ARG_FLAG(&noDirect), 			"no direct illumination",
		"-no_refWait", ARG_FLAG(&refAllLinks), 		"don't wait for all links to be refined before stopping",
		"-normalise", ARG_FLAG(&normalise), 		"rescale scene first",

// cluster + FCR

		"", 										
		"\n--- Cluster/FCR ------------------------------------------------------------\n",
		"-doFinal", ARG_FLAG(&bestPass), 			"do final 'best' pass",
		"-finLevels %d", &radOpts.bestLevels, 		"    extra levels to subdivide",
		"-finVis", ARG_FLAG(&bestVisPass), 			"    recalc visibility",
		"-noCluster", ARG_FLAG(&noClusFlag), 		"don't use clustering",
		"-scheduled", ARG_FLAG(&schedFlag), 		"use scheduled solver [default]",
		"-interleaved", ARG_FLAG(&interFlag), 		"use interleaved solver",
		"-twoStage", ARG_FLAG(&twoStageFlag), 		"use two-stage solver",
		"-schedIters %d", &radOpts.schedIterations, "max scheduled iterations to perform",


// avars

		"", 										
		"\n--- Avars ----------------------------------------------------------------------\n",
		"-set", ARG_SUBR(avarArg), 					"set avar, e.g. -set light 0.5 height 0.2",
		
// Output options
		
		"", 										
		"\n--- Output ---------------------------------------------------------------------\n",
		"-slice %F", &radOpts.sliceTime, 			"time slice: output stats at this interval",
		"-limit %F", &radOpts.limitTime, 			"time limit: halt after this much time",
		"-dumpObj", ARG_FLAG(&dumpObj), 			"output final mesh as .obj [default]",
		"-dumpSL", ARG_FLAG(&dumpSL), 				"output final mesh as .sl",
		"-updateScene", ARG_FLAG(&updateScene),		"update final mesh file at each slice",
		"-dumpScenes", ARG_FLAG(&dumpScenes),		"output separate numbered mesh file for each slice",
		"-dumpMatrix", ARG_FLAG(&dumpMatrix), 		"output form-factor matrix",
		"-dumpHier", ARG_FLAG(&dumpTree),			"output solution hierarchy",
		"-o %S", &radOpts.outFile, 					"output file name",


// Examples
		"", 										
		"\n--- Examples -------------------------------------------------------------------\n",
		"", "radiator my-scene.sl -hprog -error 0.001 -ferr 0.01",
		"", "killall -HUP  radiator               stops simulation",
		"", "killall -CONT radiator               print simulation status",
		0
	);

	// Do arg parsing
	if (argc == 1)
	{
		fprintf(stderr, "Radiator version %s\n%s\n\n%s\n\n",
				RadGetVersion().CString(),
				RadGetCopyright().CString(),
				usage.CString()
		);
				
		arg_form_print(arg_format);
		exit(0);
	}
	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

	// Process arguments
	
	if (version)
	{
		cout << "Radiator version " << RadGetVersion() << endl;
		cout << "Built on " << kBuildDate << endl;
		cout << "A radiosity engine, " << RadGetCopyright() << endl;
		exit(0);
	}

	if (!sceneFile)
	{
		cout << "No scene file specified!" << endl;
		exit(1);
	}

	radOpts.cluster = !noClusFlag;
	radOpts.bestPass = bestPass;
	radOpts.bestVisPass = bestVisPass;
	radOpts.dumpScenes = dumpScenes;	
	radOpts.updateScene = updateScene;
	radOpts.dumpTree = dumpTree;
	radOpts.drawMatrix = dumpMatrix;
	
	if (!dumpSL && !dumpScenes)
		dumpObj = true;
	

	if (ana) 
		radOpts.method = kAnalytical;
	else if (hier) 
		radOpts.method = kHierarchical;
	else if (hprog)
		radOpts.method = kProgSubstruct;
	else if (prog)
		radOpts.method = kProgressive;
	else if (mat)
		radOpts.method = kMatrix;
	else
		// default to progressive radiosity with substructuring
		radOpts.method = kProgSubstruct;
			
	if (vis44)
		radOpts.visibility = vis_4x4;
	else if (vis16)
		radOpts.visibility = vis_16x1;
	else if (vis1)
		radOpts.visibility = vis_1;
	else if (visnone)
		radOpts.visibility = vis_none;
	else
		radOpts.visibility = vis_4x4;

	if (meshRnd)
		radOpts.mesh = mesh_random;
	if (meshNL)
		radOpts.mesh = mesh_nonlin;
	radOpts.fixMesh = meshFix;
	radOpts.connectMesh = !meshNoConnect;
	radOpts.noGridMesh = meshNoGrid;

	if (!radOpts.outFile)
	// make up a name for the output file
	{
		FileName inFile;

		inFile.SetPath(sceneFile);

		if (fcrFlag)
			outFile = "fcr";
		else if (hier)
			outFile = defBasisName[radOpts.basis];
		else
			outFile = defMethodName[radOpts.method];

		outFile = outFile + "-" + inFile.GetFile();
	}
	else
		outFile = radOpts.outFile;
		
	if (frame >= 0)
	// add a frame number
		outFile = outFile + String().Printf("-%04d", frame);

	radOpts.outFile = outFile;
	// XXX make optional?
	if (true)
		LogToFile(FileName().SetPath(outFile).SetExtension("run"));
		

	if (interFlag)
		radOpts.solver = sv_interleaved;
	else if (twoStageFlag)
		radOpts.solver = sv_twoStage;
	else
		radOpts.solver = sv_scheduled;
	
	radOpts.useConjGrad = conjGrad;
	radOpts.ambient = ambient;
	radOpts.graded = !noGraded;
	radOpts.anchor = !noAnchor;
	radOpts.refAllLinks = !refAllLinks;
	radOpts.visInQuad = viq;

	radOpts.meshComplexity = meshComp;
	if (rtComp >= 0.0)
		radOpts.rtComplexity = rtComp;
	else
		radOpts.rtComplexity = meshComp;
}


Void RadMain::DoRadiosity()
{
	Int						i;
	scScenePtr				scene;
	RadMethod				*radMethod = 0;
	CreateAvarList			makeAvarList;
	FileName				path;
	
	cout << RadGetVersion() << endl;
	slInit();
	gRadControl = &radOpts;

	path.SetPath(sceneFile);
	scene = SceneReader::Load(path);

	if (!scene)
	{
		cerr << "Couldn't read file: " << sceneFile << endl;
		exit(1);
	}
	
	// Set avars...
	
		
	scene->ApplyAction(makeAvarList);

#ifdef DEBUG
	cout << "set avars:" << endl;
	for (i = 0; i < gAvarNames.NumItems(); i++)
		cout << "avar " << gAvarNames[i] << " = " << gAvarVals[i] << endl;
	cout << "scene avars:" << endl;
	for (i = 0; i < makeAvarList.avarList->NumItems(); i++)
		cout << makeAvarList.avarList->Item(i).name << " = " 
			<< makeAvarList.avarList->Item(i).value << endl;
#endif

	scene->Set(makeAvarList.avarList);
	
	for (i = 0; i < gAvarNames.NumItems(); i++)
		if (!makeAvarList.avarList->SetAvar(gAvarNames[i], gAvarVals[i]))
			cerr << "No such avar in scene: " << gAvarNames[i] 
				 << " (ignoring) " << endl;
										
	// Create radiosity method
	
	radMethod = RadMethod::NewRadMethod();

	// Run radiosity method...

	if (radMethod)
	{
		if (normalise)
			scene->Normalise();
		radMethod->SetScene(scene);
		radMethod->Render();
		if (noDirect)
			radMethod->RemoveDirect();
		
		// Output scene file
		
		if (dumpObj)
		{
			cerr << "Saving final mesh as " + radOpts.outFile + ".obj" << endl;
			radMethod->WriteObjFile(radOpts.outFile + ".obj");
		}
		if (dumpSL)
		{
			cerr << "Saving final mesh as " + radOpts.outFile + ".sl" << endl;
			radMethod->WriteSLFile(radOpts.outFile + ".sl");
		}
	}
	else
		cerr << "Sorry, that method has not been compiled in!" << endl;
}

main(int argc, char **argv)
{
	RadMain		app;

#ifdef CL_HAS_NEW
	try
	{
#endif
		app.SetOptions(argc, argv);
		signal(SIGHUP, RadHandleInterrupt);
		signal(SIGCONT, RadHandleCont);
		app.DoRadiosity();
#ifdef CL_HAS_NEW
	}
	catch (bad_alloc)
	{
		cout << "EXCEPTION: out of (virtual) memory." << endl;
		return(-1);
	}
#endif
	
	return(0);
}
