/*
	File:		RadMesh.h

	Function:	Defines a basic mesh of radiosity elements. These can be either
				quads (parallelograms), or triangles. 
					
	Author(s):	Andrew Willmott

	Copyright:	(c) 1995-2000, Andrew Willmott
 */

#ifndef __RadMesh__
#define __RadMesh__

#include "gcl/Geometry.h"
#include "gcl/Renderer.h"
#include "RadControl.h"
#include "gcl/Texture.h"
#include "cl/PtrArray.h"
#include "cl/NArray.h"

class RadElem;

class RadProps
/// Properties common to all sub-elements of an original scene polygon
{
public:
	RadProps();

	Colour				reflectance;	///< Common properties
	Colour				emittance;

	PointList			*points;		///< List of vertex positions
	ColourList			*colours;		///< List of vertex colours
	PointList			*normals;		///< List of normals 
	CoordList			*texCoords;		///< List of texture coords

	Texture				*texture;
	Int					id;

	Void				InterpolateIndexes(
							RadElem	*srcElt, 
							Int		srcVtx1, 
							Int		srcVtx2, 
							RadElem	*dstElt, 
							Int		dstVtx,
							GCLReal	m = 0.5
						);

						/**< Sets vertex dstVtx of dstElt to be the point
							 halfway between srcVtx1 and srcVtx2 on srcElt. */
};

struct CompareStats
/// Statistics comparing patches
{
	GCLReal areaSum;
	Colour	cSum;
	Colour	cSqrSum;
	Colour	refSum;
};

class PatchStats
/// Statistics for patches in the scene
{
public:
	Void			Init();
	Void			Update(RadElem *patch);
	Void			Report(ostream &s);
	
	GCLReal			minArea;
	GCLReal			maxArea;	
	Reflectance		minRefl;
	Reflectance		maxRefl;
	Colour			maxPower;
	Colour			totalPower;
};


typedef RadElem				*RadElemPtr;
typedef PtrArray<RadElem>	PatchList;
typedef Array<RadElem>		RadElemList;

class RadElem
{
public:
					RadElem();
	virtual			~RadElem();	
	
	Int  			index[4];			///< Indices of corner vertices
	Int  			clrIdx[4];			///< Indices of corner colours
	Int  			normIdx[4];			///< Indices of corner normals
	Int  			texIdx[4];			///< Indices of corner texture coords
	RadProps		*props;				///< Pointer to properties shared by
										///< elements of this mesh.
	Point 			centre;				///< centre of this elem
	Vector			normal;				///< normal at the centre
	Colour 			colour;				///< face colour -- shift this to array?
	GCLReal			area;				///< area of the quad/tri
#ifdef RAD_TEXTURE
	Colour			localRefl;
#endif
#ifdef RAD_VIS
	Byte			highlight;			///< Whether patch has been highlighted.
#endif

	virtual Void	Draw(Renderer &r);				
					// Draw this element

	//	Sampling methods

	Void			Compare(RadElem *to, GCLReal edgeLen, CompareStats &stats);
	virtual Colour	Sample(Coord c);
					///< Sample radiosity at this coord.

	Coord			FindCoord(Point &p);
					///< transform point on surface into elem's coord sys.
	virtual RadElem *FindContainer(Coord &coord);
					///< Find the elem containing this coord.
	virtual Colour	SampleLeaf(Coord c);
					///< Sample radiosity of this elem, rather than descendants.

	//	Form factor-calculation methods

	GCLReal			EstPatchFactor(const Point &p, const Vector &n);
					///< dA->dA form factor.
	GCLReal 		PatchFactor(const Point &p, const Vector &n);
					///< A->dA form factor from elem to elemental surface dS
	GCLReal			ApproxPatchFactor(const Point &p, const Vector &np);
					/**< Combines EstPatchFactor and PatchFactor according to
						 an approximation oracle */
	GCLReal			EstFormFactor(RadElem *to);
					/**< Returns an estimate of the form factor from this
						 patch to 'to' by calling one of the above
						 methods. */
	GCLReal			EdgeArea(const Vector &p, const Vector &q, 
						const Vector &n);
	GCLReal			SampledFormFactor(Int n, RadElem *to, GCLReal &error);
	Int				OrientInfo(RadElem *to);

	//	Visibility methods

	GCLReal			Visibility(RadElem *to);

	GCLReal			CentreVis(const Point &p, const Vector &n);
	GCLReal			Visibility16(const Point &p, const Vector &n);
	Bool			PotentiallyVis(RadElem *to);
	Bool			PotentiallyVisAndTouching(RadElem *to, Bool &touching);
	GCLReal			Visibility44(RadElem *to);

	Void			SetVisPoints(Point p[]);
	GCLReal 		RadVis4x4(RadElem *to);
	GCLReal 		RadVis16x1(const Point &p, const Vector &n);
	
	//	Radiosity methods....
	
	virtual Void	ColourVertices(Int weights[]);	
					///< Push element radiosities to vertices.
	virtual Void 	CreatePatches(PatchList &patches);
					///< Split Elem into patches s.t. area < maxPatchSize.
	//	Utility
	
	virtual Void 	Print(ostream &s);
	virtual Void 	Parse(istream &s);
	virtual Void	Reanimate(RadElem *parent);
	virtual StrConst Name();

	Void			SetProps(RadProps *props);
	Void			SetHighlight(Int h);
	Void			DrawHighlight(Renderer &r);
	Void			RaiseVertex(Renderer &r, Int i);
					///< Used in function-view rendering.
	Void			SetColour(const Colour &c)
					///< Set face colour.
					{ colour = c; };

	// Property accessor methods

	Vector 			&Vertex(Int i) 	{return((*props->points)[index[i]]);};
	Colour 			&VtxClr(Int i) 	{return((*props->colours)[clrIdx[i]]);};
	Vector 			&Normal(Int i) 	{return((*props->normals)[normIdx[i]]);};
	Coord 			&TexCoord(Int i) {return((*props->texCoords)[texIdx[i]]);};

	Void			SetIndexes(
						Int         dstIdx,       
						RadElem     *src,       
						Int         srcIdx
					);

	Void			SetAllIndexes(
						RadElem     *s0,       
						Int         i0,       
						RadElem     *s1,       
						Int         i1,       
						RadElem     *s2,       
						Int         i2,
						RadElem     *s3,       
						Int         i3
					);

	Vector 			&Normal() 		{return(normal);};
	Vector 			&Centre() 		{return(centre);};
	Colour 			&Emittance() 	{return(props->emittance);};
#ifdef RAD_TEXTURE
	Colour			&Reflectance()  {return(localRefl);};
#else
	Colour			&Reflectance() 	{return(props->reflectance);};
#endif
	Int				Sides()			{return((index[3] < 0) ? 3 : 4);};

	Bool			IsTri() 		{return(index[3] < 0);};
	Bool			IsQuad() 		{return(index[3] >= 0);};
	Bool			IsTextured() 	{return(props->texCoords != 0);};
	Bool			HasNormals() 	{return(props->normals != 0);};
	Void			SendPoints(Renderer &r)
					{ r.P(Vertex(0)).P(Vertex(1)).P(Vertex(2));
					  if (IsQuad()) r.P(Vertex(3)); };

	virtual GCLReal 
					MemoryUse();

	static Int		sGridMem;
	static Int		sGridChildMem;
};

ostream &operator << (ostream &s, RadElem &rq);
istream &operator >> (istream &s, RadElem &rq);

// --- An elem with neighbour info --------------------------------------------

class NbRadElem : public RadElem
{
public:
	NbRadElem();
	
	NbRadElem		*nbFace[4];		
					///< Pointers to neighbours of this patch...
	Byte			nbEdge[4];
					/** The edge of the neighbouring triangle that is adjacent
						to this one. This could be packed down into a single
						byte. */

	GCLReal			MemoryUse();
};


// --- An elem that can be gridded --------------------------------------------


class GridBase : public NbRadElem
{
public:
	Void				Mesh(GCLReal density);

	virtual RadElemPtr	AddChild(RadElem &elem, Int i, Int j, Int in, Int jn) = 0;
						///< Append a copy of the given elem to the elem list.
	Int					FindChildIndex(Coord &coord);
						///< Map from coord to the child that contains it
														
	Int					rows, cols;
};

class RadGrid : public GridBase
{
public:
	Void				CreatePatches(PatchList &patches);
	Void 				Draw(Renderer &r);
	Void				ColourVertices(Int weights[]);
	RadElem 			*FindContainer(Coord &coord);	
	Void				Reanimate(RadElem *parent);

	Void 				Print(ostream &s);
	Void 				Parse(istream &s);
	StrConst			Name();

	GCLReal				MemoryUse();

	RadElemPtr			AddChild(RadElem &elem, Int i, Int j, Int in, Int jn);	
	RadElemList			children;
};

const GCLReal kRaySurfEps = 1e-6;	// amount to offset vis testing from
									// surfaces to aVoid self-intersection.

#endif
