/*
	File:			RadMethod.h

	Function:		Defines a base class for all radiosity scene-illumination
					implementations.
					
	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott
 */


#ifndef __Rad__
#define __Rad__

#include "RadControl.h"
#include "RadMesh.h"
#include "RT_Grid.h"

#include "gcl/Scene.h"
#include "cl/Timer.h"
#include "cl/String.h"

#ifdef RAD_VIS
#include "gcl/Forms.h"
#include "gcl/ScenePane.h"
#endif

class XGraphicsSystem;

class RadAttributes
{
public:
	RadAttributes();

	ColourList	*colours;
	PointList	*points;
	VectorList	*normals;
	CoordList	*texCoords;
};

class RadMethod	: public RadAttributes	
/// Radiosity method base class.
{
public:
	RadMethod();
	virtual ~RadMethod();

	static RadMethod	*NewRadMethod();
	
	virtual Void		SetScene(scScenePtr scene);
						///< set new scene
	virtual Void		ResetOptions();
						///< called when options have been changed
	virtual Bool		Render();
						///< create the radiosity solution
	virtual Void		RemoveDirect();
						///< remove contribution from direct light sources

	Void				Print(ostream &s) const;
	Void				Parse(istream &s);
	virtual Void		PrintCmds(ostream &s) const;
	virtual Bool		ParseCmd(StrConst str, istream &s);

	virtual Void		Draw(Renderer &r);
			
	virtual Void		DrawMatrix(Renderer &r);
			
	virtual RadElem		*NewMesh() = 0;		
						///< Return mesh element for this method
	virtual PatchList	*GetElements();
						/**< Return a pointer to a list of leaf elements in the
							 mesh. */

	virtual Int			Stage(Int stage);	
						///< Gets called at various stages of solution
	virtual Bool		Idle();				
						///< For giving up time
	virtual Bool		Pause();
						///< Indicate a pause point.
	
	Void				DumpScene();
	virtual Void		DumpStats() = 0;
	Bool				CheckTime();

#ifdef RAD_VIS
	Void				RenderMatrix();

	Void				StartUpdate();
	Bool				Update();
	Void				UpdateCont();
#endif

	Void				WriteSLFile(StrConst filename);
						///< write out mesh directly as sl file
	Void				WriteObjFile(StrConst filename);
						///< write out mesh directly as wavefront obj file
	Void				WriteMRBFile(StrConst filename);

	Int					TotalMemoryUse();
						///< return total memory use in KB

// --- Raytracing -------------------------------------------------------------
	
	Void				CreateRTGrid(scScenePtr scene);
	virtual Void		CreateRTObjects(scScenePtr scene);
	Bool				IntersectsWithRay(const Point &start, const Point &end);
	virtual RadElem		*ClosestIntersection(
								const Point 	&start,
								const Vector 	&direction, 
								Point 			&hitPoint
						);

	RT_Grid				*grid;			///< ray-tracing nested grids
	RT_Object  			*object;		///< ray-traced object
	Int					rtID;

// ----------------------------------------------------------------------------

	Void				SetupMesh();
	Void				FixMesh();
	Void				ConnectMesh();
	Void				FindBounds(Point &min, Point &max);
	Void				ColourMeshInitial();
						///< Colour the patches according to initial refl/emit
	virtual Void		ColourVertices();
						///< Assign appropriate radiosities to each vertex based on face radiosities

	Void				AssignNormals();
	Void				CreatePatches();
	Void				Reanimate();
	
	Void				AddQuadTri(
							Int			numVertices,
							Int			vertices[],
							Int			changed,
							Decimator	*state
						);

	Void				SetupProps(RadProps *props, Decimator *state);

	scScenePtr			GetScene();
	
// ----------------------------------------------------------------------------

	//	fields
	PatchList			baseElems;		///< base polygons in initial mesh
	RadProps			*props;			///< properties...
	Int					numProps;
	Point				min, max;		///< scene bounds
	FindDecInfo			qtInfo;			///< number of quads/tris in the scene
		
	PatchList 			patches;		///< initial patch list
	PatchStats			stats;			///< stats about the same
	ProgramTimer		timer;			///< used to time operations
	Int					dumpID;			///< ID of last stats dump
	GCLReal				totTime;		///< time of simulation so far
	GCLReal				lastTime;		///< time of last stats dump

#ifdef RAD_VIS
	// visual stuff
	Void				SetDisplay(ScenePane *displayIn, XGraphicsSystem *gspIn)
						{ display = displayIn; gsP = gspIn; };

	XGraphicsSystem		*gsP;	
	ScenePane			*display;
	ScenePane			*matDisplay;
	Bool				doUpdate;
	GCLReal				nextUpdate;
	FL_OBJECT			*out1, *out2, *out3;
	WallClockTimer		utimer;
	PatchList			rtElemPtrs;		///< for patch selection
#endif

protected:
	scScenePtr			scene;			///< the source GCL scene
	String              sceneName;		///< name of the same
}; 


#ifdef RAD_VIS

// For the visualisation build.
#define RM_IDLE gRadControl->radObject->Idle()
#define RM_PAUSE gRadControl->radObject->Pause()
#define RM_OUT1(X) Field((gRadControl->radObject)->out1) << X << show
#define RM_OUT2(X) Field((gRadControl->radObject)->out2) << X << show
#define RM_OUT3(X) Field((gRadControl->radObject)->out3) << X << show
#define RM_DISPLAY_SCENE (gRadControl->radObject->display->Redraw())
#define RM_DISPLAY_START (gRadControl->radObject->display->Clear() \
	.Draw(gRadControl->radObject->display->ItsScene())) \
	.SetCamera(*gRadControl->radObject->display->ItsCamera())

#define RM_DISPLAY_END (gRadControl->radObject->display->Show())

#else

// For the stand-alone build
#define RM_IDLE ((RadMethod *) gRadControl->radObject)->Idle()
#define RM_PAUSE ((RadMethod *) gRadControl->radObject)->Pause()
#ifdef RAD_DEBUG
#define RM_OUT1(X) cout << "1- " << X << endl;
#define RM_OUT2(X) cout << "2- " << X << endl;
#define RM_OUT3(X) cout << "3- " << X << endl;
#else
#define RM_OUT1(X)
#define RM_OUT2(X)
#define RM_OUT3(X)
#endif
#define RM_DISPLAY_SCENE
#define RM_DISPLAY_START
#define RM_DISPLAY_END

#endif

#endif
